import unittest

from igv_reports.stream import resource_exists

class StreamTest(unittest.TestCase):


    def test_resource_exists(self):

        # File exists
        exists = resource_exists("https://raw.githubusercontent.com/igvteam/igv-data/refs/heads/main/data/test/bam/small/gstt1_sample.bam.bai")
        self.assertTrue(exists)

        # File does not exist
        exists = resource_exists("https://raw.githubusercontent.com/igvteam/igv-data/refs/heads/main/data/test/bam/small/recalibrated.bai")
        self.assertFalse(exists)

        # Host does not exist
        exists = resource_exists("https://foo.bar/none")
        self.assertFalse(exists)




from igv_reports.chralias import build_aliastable
from igv_reports.stream import getstream
from igv_reports.featureTree import FeatureTree
from igv_reports.feature import Feature

from collections import ChainMap

class WigReader:

    def __init__(self, file):
        self.file = file
        self.trees = None # use a dict of trees for each seq_header (i.e. to allow for multiple fixed/variable reagions per contig)

        self.header, self.features, self.seqnames  = parse_wig(file)
        self.trees = {k: FeatureTree(v) for k, v in self.features.items()}
        self.aliastable = build_aliastable(self.seqnames)
    
    def _query_trees(self, chrom, start, end):
        # query all trees for the region
        # return a list of features
        features = {}
        for seq, tree in self.trees.items():
            q = tree.query(chrom, start, end)
            if len(q) > 0:
                features[seq] = [f.data for f in q]
        return features

    def slice(self, region=None, region2=None):
        file_string = self.header
        features = []

        if region:
            reference = self.get_chrname(region["chr"])
            start = region["start"]
            end = region["end"]
            features.append(self._query_trees(reference, start, end))

            # Check for second region
            if region2:
                reference = self.get_chrname(region2["chr"])
                start = region2["start"]
                end = region2["end"]
                features.append(self._query_trees(reference, start, end))
        else:
            f = None
            try:
                f = getstream(self.file)
                return f.read()
            finally:
                if f:
                    f.close()
        for hdr, feat in ChainMap(*features).items():
            file_string += hdr
            for f in feat:
                file_string += f.text
        return file_string

    def get_chrname(self, c):
        if c in self.aliastable:
            return self.aliastable[c]
        else:
            return c

def parse_wig_header(header):
    # parse the key=value pairs in the header
    # return a dictionary
    header_dict = dict(ChainMap(*[{i.split('=')[0]: i.split('=')[1]} for i in header.split(' ')[1:]]))
    if 'span' not in header_dict:
        header_dict['span'] = 1
    return header_dict

def parse_wig(path):
    f = getstream(path)
    track_header = None
    step_header = None
    seq_features = {}
    seq_names = []
    current_type = None
    fixed_index = 0
    for line in f:
        if line.startswith('#'):
            continue
        elif line.startswith('track') and track_header is None:
            # this should only be on one (the first) line
            # ignores if it is redefined later in the file
            track_header = line
            continue
        elif line.startswith('fixedStep'):
            # chrom, start, step, span
            step_header_string = line
            step_header = parse_wig_header(step_header_string)
            current_type = 'fixed'
            seq_features[step_header_string] = []
            fixed_index = 0
            seq_names.append(step_header['chrom'])
            continue
        elif line.startswith('variableStep'):
            # need to get chrom and span (optional, default to 1)
            # positions are inferred from 
            step_header_string = line
            step_header = parse_wig_header(step_header_string)
            current_type = 'variable'
            seq_features[step_header_string] = []
            seq_names.append(step_header['chrom'])
            continue
        s_fields = line.strip().split()
        if current_type == 'fixed':
            start = int(step_header['start']) + (fixed_index * int(step_header['step']))
            fixed_index += 1
            end = start + int(step_header['span'])
            seq_features[step_header_string].append(Feature(step_header['chrom'], start, end, line, ''))
        elif current_type == 'variable':
            start = int(s_fields[0])
            end = start + int(step_header['span'])
            seq_features[step_header_string].append(Feature(step_header['chrom'], start, end, line, ''))
    #igv.js does not enforce the UCSC track line requirement for wig tracks
    if not track_header:
        track_header = ''
    #    raise Exception('Wig file is not properly formatted: missing track header in first line')
    return track_header, seq_features, list(set(seq_names))




def merge_regions(regions, threhold = 500):

    region_structs = []

    for r in regions:
        region_structs.append(parse_region(r))

    # build dictionary of regions by chromosome
    region_dict =  dict();
    for r in region_structs:
        chr = r['chr']
        if chr not in region_dict:
            region_dict[chr] = []
        region_dict[chr].append(r)

    merged_regions = []
    values = region_dict.values()
    for region_list in values:
        if len(region_list) == 1:
            merged_regions.append(region_list[0])
        else:
            region_list.sort(key=lambda r: r['start'])
            n = len(region_list)
            last_region = region_list[0]

            for i in range(1, n):
                reg = region_list[i]
                if reg['start'] - last_region['end'] < threhold:
                    last_region['end'] = max(last_region['end'], reg['end'])
                else:
                    merged_regions.append(last_region)
                    last_region = reg

            merged_regions.append(last_region)

    return merged_regions


def parse_region(locus_string):

    tokens = locus_string.split(":")
    chr = tokens[0]

    t2 = tokens[1].split('-')
    start = int(t2[0].replace(',', ''))
    if len(t2) > 1:
        end = int(t2[1].replace(',',''))
    else:
        end = start

    return {'chr': chr, 'start': start, 'end': end}


import unittest

from igv_reports.vcf import VcfReader
import pathlib

def get_data(path, region = None):
    vcf = VcfReader(path)
    return vcf.slice(region)


class VcfTest(unittest.TestCase):

    def test_getvariants(self):

        path = str((pathlib.Path(__file__).parent / "data/minigenome/variants.vcf").resolve())
        region = {"chr": "minigenome", "start": 87944 - 1, "end": 87944 + 1}
        data = get_data(path, region)
        count = count_variants(data)
        self.assertEqual(count, 1)

        path = str((pathlib.Path(__file__).parent / "data/minigenome/variants.vcf.gz").resolve())
        region = {"chr": "minigenome", "start": 87944 - 1, "end": 87944 + 1}
        data = get_data(path, region)
        count = count_variants(data)
        self.assertEqual(count, 1)

        path = str((pathlib.Path(__file__).parent / "data/minigenome/variants.bcf").resolve())
        region = {"chr": "minigenome", "start": 87944 - 1, "end": 87944 + 1}
        data = get_data(path, region)
        count = count_variants(data)
        self.assertEqual(count, 1)

    def test_remote(self):

        path = "https://raw.githubusercontent.com/igvteam/igv-reports/refs/heads/main/reports/variants/cancer.vcf.gz"
        region = {"chr": "chr17", "start": 7673767, "end": 43071077}
        data = get_data(path, region)
        count = count_variants(data)
        self.assertTrue(count > 0)

    def test_multilocus(self):
        path = str((pathlib.Path(__file__).parent / "../test/data/variants/SKBR3_Sniffles_variants_tra.vcf").resolve())
        vcf = VcfReader(path)
        region = {"chr": "8", "start": 77706277-1, "end": 77706277+1}
        region2 = {"chr": "14", "start": 50256582-1, "end": 50256582+1}
        data = vcf.slice(region, region2)
        count = count_variants(data)
        self.assertEqual(count, 2)

    def test_alias(self):
        path = str((pathlib.Path(__file__).parent / "../test/data/variants/SKBR3_Sniffles_variants_tra.vcf").resolve())
        vcf = VcfReader(path)
        region = {"chr": "chr8", "start": 77706277-1, "end": 77706277+1}
        region2 = {"chr": "chr14", "start": 50256582-1, "end": 50256582+1}
        data = vcf.slice(region, region2)
        count = count_variants(data)
        self.assertEqual(count, 2)

    def test_missingchr(self):
        path = str((pathlib.Path(__file__).parent / "../test/data/variants/SKBR3_Sniffles_variants_tra.vcf").resolve())
        vcf = VcfReader(path)
        region = {"chr": "99", "start": 77706277-1, "end": 77706277+1}
        data = vcf.slice(region)
        count = count_variants(data)
        self.assertEqual(count, 0)

    def test_notindexed(self):
        path = str((pathlib.Path(__file__).parent / "../test/data/variants/1kg_phase3_sites.vcf").resolve())
        vcf = VcfReader(path)
        region = {"chr": "chr22", "start": 50173573-1, "end": 50173573+1}
        data = vcf.slice(region)
        count = count_variants(data)
        self.assertEqual(2, count)


    def test_tabix(self):
        path = str((pathlib.Path(__file__).parent / "../test/data/variants/1kg_phase3_sites.vcf.gz").resolve())
        vcf = VcfReader(path)
        region = {"chr": "chr22", "start": 50173573-1, "end": 50173573}
        data = vcf.slice(region)
        count = count_variants(data)
        self.assertEqual(2, count)

    def test_tabix_alias(self):
        path = str((pathlib.Path(__file__).parent / "../test/data/variants/1kg_phase3_sites.vcf.gz").resolve())
        vcf = VcfReader(path)
        region = {"chr": "22", "start": 50173573-1, "end": 50173573}
        data = vcf.slice(region)
        count = count_variants(data)
        self.assertEqual(2, count)

    def test_tabix_missingchr(self):
        path = str((pathlib.Path(__file__).parent / "../test/data/variants/1kg_phase3_sites.vcf.gz").resolve())
        vcf = VcfReader(path)
        region = {"chr": "99", "start": 50173573-1, "end": 50173573}
        data = vcf.slice(region)
        count = count_variants(data)
        self.assertEqual(count, 0)


def count_variants(data):

    lines = data.split('\n')
    count = 0
    header_read = False
    for line in lines:
        if header_read and len(line) > 0:
            print(line)
            count+=1
        else:
            if line.startswith("#CHROM"):
                header_read = True
    return count



from igv_reports import feature

def get_track_json_dict(url):

    name = get_name(url)
    format = feature.infer_format(url)
    type = get_track_type(format)

    trackobj = {
        "name": name,
        "url": url,
        "type": type,
        "format": format
    }

    if type == "alignment":
        trackobj["height"] = 500
    elif type == "mut":
        trackobj["height"] = 50
        trackobj["color"] = "rgb(0,0,150)"

    # To respect --exclude-flags, show all reads which are already filtered by
    # pysam in bam.py. igv.js defaults duplicate and vendorFailed to True.
    if format == "bam" or format == "cram":
        trackobj["filter"] = {"duplicate": False, "vendorFailed": False}
    return trackobj

def get_name(filename):

    idx = filename.rfind("/")
    if idx < 0:
        idx = filename.rfind("\\")
    period = filename.rfind(".")
    if idx < 0:
        if period < 0:
            return filename
        else:
            return filename[:period]
    else:
        idx += 1
        if period < 0:
            return filename[idx]
        else:
            return filename[idx:period]


def get_track_type(format):

    dict = {
        "bam": "alignment",
        "cram": "alignment",
        "bed": "annotation",
        "gff3": "annotation",
        "gff": "annotation",
        "gtf": "annotation",
        "bed": "annotation",
        "refgene": "annotation",
        "bcf": "variant",
        "vcf": "variant",
        "wig": "wig",
        "bedgraph": "wig",
        "maf": "mut"
    }
    return dict[format] if format in dict else None

def is_format_supported(format):
    return get_track_type(format) is not None



import unittest
import pathlib
from igv_reports.ideogram import IdeogramReader

class IdeogramTest(unittest.TestCase):

    def test_extract_chr(self):
        data_path = str((pathlib.Path(__file__).parent / 'data/hg19/cytoBandIdeo.txt').resolve())
        ideogram_reader = IdeogramReader(data_path)
        extracted = ideogram_reader.get_data("chr1")
        self.assertTrue(extracted.startswith('chr1'))

    def test_extract_chralias(self):
        data_path = str((pathlib.Path(__file__).parent / 'data/hg19/cytoBandIdeo.txt').resolve())
        ideogram_reader = IdeogramReader(data_path)
        extracted = ideogram_reader.get_data("2")
        self.assertTrue(extracted.startswith('chr2'))

    def test_multilocuas(self):

        data_path = str((pathlib.Path(__file__).parent / 'data/hg19/cytoBandIdeo.txt').resolve())
        ideogram_reader = IdeogramReader(data_path)
        ideo_string = ideogram_reader.get_data("chr18")
        ideo_string += ideogram_reader.get_data("chr19")




import unittest

from igv_reports.utils import resolve_relative_path

class UtilsTest(unittest.TestCase):


    def test_resolve_relative_path(self):

        # Relative path
        f = resolve_relative_path("/Users/foo/bar", "fn")
        self.assertEqual("/Users/foo/fn", f)

        # Absolute path
        f = resolve_relative_path("/Users/foo/bar", "/a/b/fn")
        self.assertEqual("/a/b/fn", f)

        # URL relative path
        f = resolve_relative_path("https://Users/foo/bar?q=p", "fn")
        self.assertEqual("https://Users/foo/fn?q=p", f)

        # URL absolute path
        f = resolve_relative_path("https://Users/foo/bar?q=p", "https://a/b/fn")
        self.assertEqual("https://a/b/fn", f)


from os import path
from .stream import getstream
from . import feature, bam, vcf, wig

def getreader(config, filetype=None, args = None):

    path = config["url"]
    indexed = config["indexed"] if "indexed" in config else True    # Assume true until proven otherwise

    if not filetype:
        filetype = feature.infer_format(path)
    else:
        filetype = filetype.lower()

    if filetype == "bam" or filetype == "cram" :
        return bam.BamReader(filetype, path, args)

    elif filetype == "vcf" or filetype == "bcf":
        return vcf.VcfReader(path)
    
    elif filetype == "wig":
        return wig.WigReader(path)

    else:
        return feature.get_featurereader(config)



def resolve_relative_path(path1, path2):

    if path2.startswith("http://") or path2.startswith("https://") or path.isabs(path2):
        return path2

    if path1.startswith("http://") or path1.startswith("https://"):
        qIdx = path1.find("?")
        strippedURL = path1[0:qIdx] if qIdx > 0 else path1
        query = path1[qIdx::] if qIdx > 0 else ""
        lastSlashIdx  = len(strippedURL) - strippedURL[::-1].find("/")
        return strippedURL[0:lastSlashIdx] + path2 + query

    else:
        if path.exists(path2):
            return path2
        else:
            dir = path.dirname(path1)
            return path.join(dir, path2)


def get_content_as_utf8(source):
    """
    Reads the content from a local file or a URL and returns it as a UTF-8 string.

    Args:
        source (str): The local file path or URL.

    Returns:
        str: The content as a UTF-8 string.
    """
    f = getstream(source)
    return f.read()



import requests
import json



def get_genome(id):

    genomes_url = "https://igv.org/genomes/genomes3.json"
    r = requests.get(genomes_url)

    if r.status_code == 200:
        genomes = r.json()
        genome_ids = []
        for g in genomes:
            if g["id"] == id:
                return g
            genome_ids.append(g["id"])

        # genome not found
        msg = f'Unknown genome ID: {id}. Valid genome values: {", ".join(genome_ids)}'
        raise ValueError(msg)

    else:
        print(f'Error loading genomes {r.status_code}')
        return None



def main():
    genome = get_genome("hg38")
    print (genome)

if __name__ == "__main__":
    main()



import pysam
from igv_reports.featureTree import FeatureTree
from igv_reports.chralias import build_aliastable, get_alias


class VcfReader:

    def __init__(self, path):

        self.file = pysam.VariantFile(path)
        self.feature_tree = None

        if self.file.index is None:
            records = self.file.fetch()
            nodes = list(map(lambda rec: TreeNode(rec.contig, rec.pos - 1, rec.stop, rec), records))
            self.feature_tree = FeatureTree(nodes)
            chrs = self.feature_tree.getchrs()
            self.aliastable = build_aliastable(chrs)

        else:
            # Build chr alias table from Contig lines in file header.  These are optional, but are usually there
            chrs = get_contigs(str(self.file.header))
            self.aliastable = build_aliastable(chrs)

    def slice(self, region=None, region2=None):

        vcf = self.file

        header = vcf.header

        fileString = str(header)

        if region == None:
            records = vcf.fetch()
        else:
            records = self.fetch(self.get_chrname(region['chr']), region['start'], region['end'])

            # Check for optional second region
            if region2 is not None:
                records2 = self.fetch(self.get_chrname(region2['chr']), region2['start'], region2['end'])
                for rec in records2:
                    records.append(rec)

        for rec in records:
            fileString += (str(rec))

        return fileString

    '''
    Wrap pysam fetch with workaround for non-indexed files. This exact logic is repeated in feature.py.
    '''

    def fetch(self, chr, start, end):

        if chr is None:
            return []

        vcf = self.file
        if vcf.index is not None:
            try:
                return list(vcf.fetch(chr, start, end))
            except ValueError:

                # Possible chr alias error (e.g. 1 vs chr1)
                try:
                    alias = get_alias(chr)
                    value = vcf.fetch(alias, start, end)
                    self.aliastable[chr] = alias
                    return value
                except ValueError:
                    self.aliastable[chr] = None
                    print(f'WARNING: No data with sequence name {chr} found in file {vcf.filename}')
                    return []

        else:
            nodes = self.feature_tree.query(chr, start, end)
            return list(filter(lambda rec: rec.stop >= start and rec.start <= end, map(lambda node: node.data.record, nodes)))

    def get_chrname(self, c):
        if c in self.aliastable:
            return self.aliastable[c]
        else:
            return c


class TreeNode:

    def __init__(self, chr, start, end, record):
        self.chr = chr
        self.start = start
        self.end = end
        self.record = record


def parse_info(record):
    info_str = str(record).split('\t')[7]
    fields = info_str.split(';')
    info_dict = {}
    for f in fields:
        kv = f.split('=')
        if len(kv) == 2:
            info_dict[kv[0]] = kv[1]
    return info_dict


def get_contigs(data):
    contigs = []
    lines = data.split('\n')
    for line in lines:
        if line.startswith('##contig'):
            idx1 = line.find("ID=")
            if idx1 > 0:
                idx2 = line.find(",")
                contigs.append(line[idx1 + 3:idx2])

    return contigs



import json
import sys
import html
from .feature import parse


class GenericTable:

    @staticmethod
    def from_tabfile(file, info_columns=None, sequence=None, begin=None, end=None, zero_based=False):

        if sequence is not None and begin is not None and end is not None:
            seq_col = int(sequence) - 1
            start_col = int(begin) - 1
            end_col = int(end) - 1
        else:
            presets = determine_presets(file)
            seq_col = presets[0]
            start_col = presets[1]
            end_col = presets[2]
            zero_based = presets[3]

        if start_col < 0:
            msg = f"Invalid begin value: {begin}"
            sys.exit(msg)
        if end_col < 0:
            msg = f"Invalid end value: {end}"
            sys.exit(msg)

        regions = []
        rows = parse(file, 'tab')
        unique_id = 0
        start_offset = 0 if zero_based else 1

        # set header to be first row and assume file has a header
        header = rows[0]
        rows.pop(0)
        
        for row in rows:
            chr = row[seq_col]
            start = int(row[start_col]) - start_offset
            end = int(row[end_col])
            regions.append((_Region(chr, start, end), unique_id))
            unique_id += 1

        return GenericTable(rows, regions, info_columns, header)

    @staticmethod
    def from_fusionjson(file):

        """
        Fusion json files have the following properties.  Not all are present
        {
            "# Long Reads": "305",
            "Annotations": "[\"INTRACHROMOSOMAL[chr11:0.01Mb]\",\"NEIGHBORS_OVERLAP:+:-:[5797]\"]",
            "Fusion": "BEST1--FTH1",
            "Left Breakpoint": "chr11:61965476:+",
            "Right Breakpoint": "chr11:61967627:-",
            "Splice Type": "INCL_NON_REF_SPLICE"
        }
        """

        with open(file) as json_file:

            j = json.load(json_file)

            header = j["columnns"] if "columns" in j else ['Fusion', "# Long Reads", 'Junction Reads',
                                                           'Spanning Fragments', "Expr Level (FFPM)", 'Splice Type',
                                                           'Left Gene', 'Left Breakpoint', 'Right Breakpoint',
                                                           "Annotations"]
            unique_id = 0
            features = []
            rows = []
            for record in j["fusions"]:
                row = []
                for h in header:
                    v = record[h] if h in record else ''
                    row.append(v)
                rows.append(row)

                chr = record["Fusion"]
                features.append((_Region(chr, None, None), unique_id))
                unique_id = unique_id + 1

        return GenericTable(rows, features, header, header)

    def __init__(self, rows, features, info_columns, header):

        self.column_names = info_columns
        self.features = features
        self.header = header
        self.rows = rows

    def to_JSON(self):

        obj = {
            "headers": ["unique_id"],
            "rows": []
        }

        if self.column_names is None:
            indeces = range(0, len(self.header))
        else:
            indeces = []
            for h in self.column_names:
                try:
                    idx = self.header.index(h)
                    indeces.append(idx)
                except:
                    print(f"Column {h} is not present")

        for i in indeces:
            obj["headers"].append(html.escape(self.header[i]))

        unique_id = 0
        for row in self.rows:
            r = [unique_id]
            for i in indeces:
                value = row[i] if i < len(row) else ""
                r.append(html.escape(value))
            obj["rows"].append(r)
            unique_id += 1

        return json.dumps(obj)


def determine_presets(filename):
    filename = filename.lower()
    if (filename.endswith(".gz")):
        filename = filename[:-3]
    if filename.endswith(".bed"):
        return (0, 1, 2, True)
    elif filename.endswith(".gff") or filename.endswith("gff3") or filename.endswith("gtf"):
        return (0, 3, 4, False)
    elif filename.endswith(".maf"):
        return (4, 5, 6, False)
    elif filename.endswith(".mut"):
        return (0, 1, 2, False)
    else:
        msg = f"Unknown file format for: {filename} Please set --sequence, --start, and --end"
        sys.exit(msg)


class _Region:

    def __init__(self, chr, start = None, end = None):
        self.chr = chr
        self.start = start
        self.end = end



import argparse
import json
import math
import os
import sys
from urllib.request import urlopen

import yaml

from igv_reports import datauri, tracks, utils, feature
from igv_reports.bedtable import BedTable
from igv_reports.bedtable import BedpeTable
from igv_reports.bedtable import JunctionBedTable
from igv_reports.fasta import FastaReader
from igv_reports.generictable import GenericTable
from igv_reports.genome import get_genome
from igv_reports.ideogram import IdeogramReader
from igv_reports.regions import parse_region
from igv_reports.stream import resource_exists
from igv_reports.tracks import get_track_type, is_format_supported
from igv_reports.twobit import TwoBitReader
from igv_reports.utils import resolve_relative_path
from igv_reports.varianttable import VariantTable

'''
Create an html report.  This is the main function for the application.
'''


def create_report(args):
    # Read filter config if provided
    filter_config = None
    if args.tabulator and args.filter_config:
        try:
            with open(args.filter_config, 'r') as f:
                filter_config = yaml.safe_load(f)
        except Exception as e:
            print(f"Error reading filter config: {e}")

    # Read the variant data -- this populates the variant table and defines the corresponding regions

    variants_file = args.sites

    if variants_file.endswith(".bcf") or variants_file.endswith(".vcf") or variants_file.endswith(".vcf.gz"):
        table = VariantTable(variants_file, args)

    elif variants_file.endswith(".bed") or variants_file.endswith(".bed.gz"):
        if args.type is not None and args.type == "junction":
            table = JunctionBedTable(variants_file, args.info_columns)
        else:
            table = BedTable(variants_file)

    elif variants_file.endswith(".bedpe") or variants_file.endswith(".bedpe.gz"):
        table = BedpeTable(variants_file)

    elif variants_file.endswith(".maf") or variants_file.endswith(".maf.gz") or (
            args.sequence is not None and args.begin is not None and args.end is not None):
        table = GenericTable.from_tabfile(variants_file, args.info_columns, args.sequence, args.begin, args.end,
                                          args.zero_based)

    elif variants_file.endswith(".json"):
        table = GenericTable.from_fusionjson(variants_file)

    # Track json array.  Tracks can come from (1) --tracks argument, (2) --genome  argument, and (3) --track_config argument
    trackjson = []

    # Check for optional genome argument.
    if args.genome is not None:
        genome = get_genome(args.genome)

        # Sequence - prefer twobit if defined
        if args.twobit is None and "twoBitURL" in genome:
            args.twobit = genome["twoBitURL"]
        elif args.fasta is None and "fastaURL" in genome:
            args.fasta = genome["fastaURL"]

        # Ideogram
        if args.ideogram is None and "cytobandURL" in genome:
            args.ideogram = genome["cytobandURL"]

        # Tracks
        if "tracks" in genome:
            for config in genome["tracks"]:
                if "format" not in config and "url in c":
                    config["format"] = feature.infer_format(config["url"])
                if is_format_supported(config["format"]):
                    if "type" not in config:
                        config["type"] = get_track_type(config["format"])
                    trackjson.append(config)
                else:
                    print("File format: " + config["format"] + " is not supported. Skipping track '" + config[
                        "name"] + "'.")

    # --tracks argument
    if args.tracks is not None:
        for track in args.tracks:
            config = tracks.get_track_json_dict(track)
            # If this is a no-embed report add index URLs
            if args.no_embed == True:
                add_index(config)
            trackjson.append(config)

    # --roi argument
    if args.roi is not None:
        for track in args.roi:
            config = tracks.get_track_json_dict(track)
            config["type"] = "roi"
            # If this is a no-embed report add index URLs
            if args.no_embed == True:
                add_index(config)
            trackjson.append(config)

    # --track_config argument
    if args.track_config is not None:
        for trackobj in args.track_config:
            with open(trackobj) as f:
                j = json.load(f)
                for config in j:
                    if "url" in config:
                        config["url"] = resolve_relative_path(trackobj, config["url"])
                    if "format" not in config and "url" in config:
                        config["format"] = feature.infer_format(config["url"])
                    if "type" not in config:
                        config["type"] = get_track_type(config["format"])
                    trackjson.append(config)

    if args.no_embed == True:
        igv_config = json.dumps(create_noembed_session(args, trackjson))
        locus_dict = json.dumps(create_locus_dict(table, args.window))

    # Create the session dictionary json, containing a session object for each variant
    else:
        session_dict = json.dumps(create_session_dict(args, table, trackjson, args.sampleinfo))

    # Generate the HTML
    template_file = args.template
    if None == template_file:
        if args.type == "junction":
            template_file = os.path.dirname(sys.modules['igv_reports'].__file__) + '/templates/junction_template.html'
        elif args.type == "fusion":
            template_file = os.path.dirname(sys.modules['igv_reports'].__file__) + '/templates/fusion_template.html'
        elif args.no_embed:
            template_file = os.path.dirname(
                sys.modules['igv_reports'].__file__) + '/templates/variant_template-noembed.html'
        elif args.tabulator:
            template_file = os.path.dirname(
                sys.modules['igv_reports'].__file__) + '/templates/variant_template_tabulator.html'
        else:
            template_file = os.path.dirname(sys.modules['igv_reports'].__file__) + '/templates/variant_template.html'

    output_file = args.output

    standalone = args.standalone
    with open(template_file, "r") as f:
        data = f.readlines()

        with open(output_file, "w") as o:

            for i, line in enumerate(data):

                if standalone:
                    if line.strip().startswith("<script") and ".js\"" in line:
                        inline_script(line, o, "js")
                        continue
                    elif line.strip().startswith("<link") and line.strip().endswith("css\">"):
                        inline_script(line, o, "css")
                        continue

                if args.title and line.strip().startswith("<title>"):
                    line = "<title>" + args.title + "</title>"

                if args.header and line.strip().startswith("<!--header-->"):
                    line = "<div>" + read_contents(args.header) + "</div>"

                if args.footer and line.strip().startswith("<!--footer-->"):
                    line = "<div>" + read_contents(args.footer) + "</div>"

                j = line.find('"@TABLE_JSON@"')
                if j >= 0:
                    line = line.replace('"@TABLE_JSON@"', table.to_JSON())

                j = line.find('"@SESSION_DICTIONARY@"')
                if j >= 0:
                    line = line.replace('"@SESSION_DICTIONARY@"', session_dict)

                j = line.find('"@FILTER_CONFIG@"')
                if j >= 0:
                    filter_config_json = json.dumps(filter_config) if filter_config else "null"
                    line = line.replace('"@FILTER_CONFIG@"', filter_config_json)

                j = line.find('"@LOCUS_DICTIONARY@"')
                if j >= 0:
                    line = line.replace('"@LOCUS_DICTIONARY@"', locus_dict)

                j = line.find('"@IGV_CONFIG@"')
                if j >= 0:
                    line = line.replace('"@IGV_CONFIG@"', igv_config)

                sort_option = args.sort.upper() if args.sort is not None else "BASE"
                j = line.find('"@SORT@"')
                if j >= 0:
                    line = line.replace('"@SORT@"', '"' + sort_option + '"')

                o.write(line)


def create_session_dict(args, table, trackjson, sampleinfo):
    ''' Create a dictionary of igv.js session objects, one for each variant '''

    session_dict = {}

    # Create file readers for tracks.  This is done outside the locus loop so initialization happens once
    readers = []
    for config in trackjson:
        readers.append(utils.getreader(config, None, args))

    # Other readers
    if args.fasta is not None:
        sequence_reader = FastaReader(args.fasta)
    elif args.twobit is not None:
        sequence_reader = TwoBitReader(args.twobit)
    else:
        raise ValueError('Must specify either fasta or twobit')

    if args.ideogram is not None:
        ideogram_reader = IdeogramReader(args.ideogram)
    else:
        ideogram_reader = None

    # loop through regions defined from variant, annotation, or bedpe files,  creating an igv.js session for each one
    flanking = 1000
    if args.flanking is not None:
        flanking = float(args.flanking)
    i = 0
    for tuple in table.features:
        i += 1
        print(f"Working on variant {i}/{len(table.features)}")

        feature = tuple[0]
        unique_id = tuple[1]

        # If a variant feature (=> row in the table) has an explicit session id use it, otherwise use the row id (i.e. unique_id)
        if hasattr(feature, "session_id"):
            session_id = feature.session_id
        else:
            session_id = str(unique_id)

        if session_id not in session_dict:

            # Placeholder variable for possible second region (bedpe files)
            region2 = None

            # Define a genomic region around the variant
            if hasattr(feature, "viewport"):
                region = parse_region(feature.viewport)
                chr = region["chr"]
                start = region["start"]
                end = region["end"]
            else:
                chr = feature.chr

                start = max(1, int(math.floor(feature.start - flanking / 2))) if feature.start is not None else None
                end = int(math.ceil(feature.end + flanking / 2)) if feature.end is not None else None

                region = {"chr": chr, "start": start, "end": end}
                region2 = None

                # If feature has a second locus (bedpe file) create the region here.
                if hasattr(feature, 'chr2') and feature.chr2 is not None:
                    chr2 = feature.chr2
                    start2 = max(1, int(math.floor(feature.start2 - flanking / 2))) if feature.start2 is not None else None
                    end2 = int(math.ceil(feature.end2 + flanking / 2)) if feature.end2 is not None else None

                    # If regions are on the same chromosome and overlap, merge them.  This is for data storage, initial locus is set elsewhere.
                    if chr2 == chr and all(v is not None for v in [start, end, start2, end2]) and start <= end2 and end >= start2:
                        start, end = min(start, start2), max(end, end2)
                        region = {"chr": chr, "start": start, "end": end}
                    else:
                        region2 = {"chr": chr2, "start": start2, "end": end2}

            # Sequence
            data = sequence_reader.slice(region)
            length = sequence_reader.get_reference_length(chr)
            fa = '>' + chr + ':' + str(start) + '-' + str(end) + ' @len=' + str(length) + '\n' + data

            if region2 is not None:
                data2 = sequence_reader.slice(region2)
                length2 = sequence_reader.get_reference_length(chr2)
                fa += '\n' + '>' + chr2 + ':' + str(start2) + '-' + str(end2) + ' @len=' + str(length2) + '\n' + data2

            fasta_uri = datauri.get_data_uri(fa)
            fastaJson = {
                "fastaURL": fasta_uri,
            }

            # Ideogram
            if args.ideogram is not None:
                ideo_string = ideogram_reader.get_data(region["chr"])
                if region2 is not None:
                    ideo_string += ideogram_reader.get_data(region2["chr"])
                ideo_uri = datauri.get_data_uri(ideo_string)
                fastaJson["cytobandURL"] = ideo_uri

            # Initial locus  -- ths is based on the feature or viewport, not the region which includes flanking sequence
            if (hasattr(feature, "viewport")):
                initial_locus = feature.viewport
            else:
                if not hasattr(feature, 'chr2') or feature.chr2 is None:
                    # Single locus feature
                    initial_locus = locus_string(feature.chr, feature.start, feature.end, args.window)
                else:
                    # Bedpe feature with two loci
                    locus1 = locus_string(feature.chr, feature.start, feature.end, args.window)
                    locus2 = locus_string(feature.chr2, feature.start2, feature.end2, args.window)

                    # Check for overlap to determine if loci should be merged or concatenated
                    if args.merge_overlaps and feature.chr2 == feature.chr and feature.start <= feature.end2 and feature.end >= feature.start2:
                        merged_start = min(feature.start, feature.start2)
                        merged_end = max(feature.end, feature.end2)
                        initial_locus = locus_string(feature.chr, merged_start, merged_end, args.window)
                    else:
                        initial_locus = f'{locus1} {locus2}'


            # Loop through track configs
            tracks = []
            if args.translate_sequence_track:
                tracks.append({"type": "sequence", "frameTranslate": True})
            roi = []
            track_order = 1
            idx = 0
            for config in trackjson:

                reader = readers[idx]
                idx += 1

                # "cram" input format is converted to "bam" for output track configs
                if config["format"] == "cram":
                    config["format"] = "bam"

                if config["format"] == "bcf":
                    config["format"] = "vcf"

                # Indexes are not used with data URIs
                if "indexURL" in config:
                    del config["indexURL"]

                data = reader.slice(region, region2)

                # replace url with data URI
                config["url"] = datauri.get_data_uri(data)

                if (config["type"] == "alignment"):
                    if "height" not in config:
                        config["height"] = 500

                    is_snv = feature.end is not None and feature.end - feature.start == 1

                    if (config["type"]) == "alignment" and (args.sort is not None or is_snv) and (
                            args.sort != 'NONE'):
                        sort_option = 'BASE' if args.sort is None else args.sort.upper()
                        config["sort"] = {
                            "option": sort_option,
                            "chr": chr,
                            "position": str(feature.start + 1),
                            "direction": "ASC"
                        }
                if "order" not in config:
                    config["order"] = track_order
                track_order += 1

                if "roi" == config["type"]:
                    roi.append(config)
                else:
                    tracks.append(config)

            session_json = {
                "locus": initial_locus,
                "reference": fastaJson,
                "tracks": tracks
            }
            if len(roi) > 0:
                session_json["roi"] = roi

            # --sampleinfo argument
            if args.sampleinfo is not None:
                sampleinfoJson = []
                for si in args.sampleinfo:
                    data = utils.get_content_as_utf8(si)
                    sampleinfoJson.append({"url": datauri.get_data_uri(data)})
                session_json["sampleinfo"] = sampleinfoJson

            # Build the session data URI
            session_string = json.dumps(session_json)
            session_uri = datauri.get_data_uri(session_string)
            session_dict[session_id] = session_uri

    return session_dict


# Create an igv session config object.  This is used for no-embed reports
def create_noembed_session(args, trackjson):
    if args.fasta is not None:
        reference = {
            "fastaURL": args.fasta
        }
        if resource_exists(args.fasta + ".fai"):
            reference["indexURL"] = args.fasta + ".fai"

    elif args.twobit is not None:
        reference = {
            "twoBitURL": args.twobit
        }
    else:
        raise ValueError('Must specify either fasta or twobit')

    if args.ideogram is not None:
        reference["cytobandURL"] = args.ideogram

    tracks = []
    roi = []
    for tj in trackjson:
        if "roi" == tj["type"]:
            roi.append(tj)
        else:
            tracks.append(tj)
    session = {
        "reference": reference,
        "tracks": tracks
    }
    if len(roi) > 0:
        session["roi"] = roi

    if args.sampleinfo is not None:
        sampleinfoJson = []
        for si in args.sampleinfo:
            sampleinfoJson.append({"url": si})
        session["sampleinfo"] = sampleinfoJson

    return session


def create_locus_dict(table, window):
    locus_dict = {}

    # loop through regions defined from variant, annotation, or bedpe files,  creating  locus for each one

    for tuple in table.features:

        feature = tuple[0]
        unique_id = tuple[1]

        if hasattr(feature, "session_id"):
            session_id = feature.session_id
        else:
            session_id = str(unique_id)

        if (hasattr(feature, "viewport")):
            locus = feature.viewport
        else:
            locus = locus_string(feature.chr, feature.start + 1, feature.end, window)
            if hasattr(feature, 'chr2') and feature.chr2 is not None:
                locus += f' {locus_string(feature.chr2, feature.start2, feature.end2, window)}'

        locus_dict[session_id] = locus

    return locus_dict


def inline_script(line, o, source_type):
    # <script type="text/javascript" src="https://igv.org/web/test/dist/igv.min.js"></script>
    if source_type == "js":
        s = line.find('src="')
        offset = 5
        o.write('<script type="text/javascript">\n')
    elif source_type == "css":
        s = line.find('href="')
        offset = 6
        o.write('<style type="text/css">\n')
    else:
        raise KeyError("Inline script must be either js- or css-file")
    if s > 0:
        e = line.find('">', s)
        url = line[s + offset:e]
        response = urlopen(url)
        content = response.read().decode('utf-8')
        response.close()
        o.write(content)
        if source_type == "js":
            o.write('</script>\n')
        else:
            o.write('</style>\n')
    else:
        raise ValueError("No file path in {l} for inline script.".format(l=line))


def locus_string(chr, start, end, window):
    if start is None:
        return chr

    if window is not None:
        window = int(window)
        return f'{chr}:{start + 1 - window / 2}-{end + window / 2}'
    else:
        if (end - start) == 1:
            return f'{chr}:{start + 1}'
        else:
            return f'{chr}:{start + 1}-{end}'


# Potentially add an index URL to a track config.  The "format" field must be set before calling this function
def add_index(config):
    if "url" not in config or "indexURL" in config:
        return

    indexURL = None
    url = config["url"]

    # Check tabix first
    if url.endswith(".gz"):
        if resource_exists(url + ".tbi"):
            indexURL = url + ".tbi"
        elif resource_exists(url + ".csi"):
            indexURL = url + ".csi"

    # Check potential bam/cram files
    if "format" in config:
        format = config["format"]
        if format == "bam":
            if resource_exists(url + ".bai"):
                indexURL = url + ".bai"
            elif resource_exists(format + ".csi"):
                indexURL = url + ".csi"
            else:
                k = url.rfind(".")
                tmp = url[:k] + ".bai"
                if resource_exists(tmp):
                    indexURL = tmp
        elif format == "cram" and indexURL == None:
            if (resource_exists(url + ".crai")):
                indexURL = url + ".crai"

    if indexURL is not None:
        config["indexURL"] = indexURL


def read_contents(file_path):
    """Reads the contents of a file and returns it as a string."""
    if not os.path.exists(file_path):
        raise FileNotFoundError("File not found: " + file_path)
    with open(file_path, 'r') as file:
        return file.read()


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("sites", help="vcf file defining variants, required")
    parser.add_argument("fasta_", nargs="?", default=None,
                        help="reference fasta file.  Deprecated positional argument,  use --fasta")
    parser.add_argument("--fasta", nargs="?", default=None,
                        help="reference fasta file. One of either --fasta, --twobit, or --genome is required.")
    parser.add_argument("--twobit", nargs="?", default=None,
                        help="reference twobit file. One of either --fasta, --twobit, or --genome is required.")
    parser.add_argument("--genome", help="igv.js genome id (e.g. hg38)")

    parser.add_argument("--type",
                        help="Report type.  Possible values are mutation, junction, and fusion.  Default is mutation")
    parser.add_argument("--ideogram", help="ideogram file in UCSC cytoIdeo format")
    parser.add_argument("--tracks", nargs="+", help="list of track files")
    parser.add_argument("--track-config", nargs="+", help="track json file")
    parser.add_argument("--roi", nargs="+", help="list of region-of-interest files")
    parser.add_argument("--sort",
                        help="initial sort option for alignment tracks.   Supported values include  BASE, STRAND, INSERT_SIZE, and MATE_CHR. Default value is BASE for single nucleotide variants, no sorting otherwise.  See the igv.js documentation for more information. ")
    parser.add_argument("--template", help="html template file", default=None)
    parser.add_argument("--output", help="output file name", default="igvjs_viewer.html")
    parser.add_argument("--info-columns", nargs="+", help="list of VCF info field names to include in variant table")
    parser.add_argument("--info-columns-prefixes", nargs="+",
                        help="list of prefixes of VCF info field names to include in variant table")
    parser.add_argument("--sampleinfo", nargs="+", help="list of sample information files")
    parser.add_argument("--samples", nargs="+",
                        help="Space delimited list of sample (i.e. genotypes) names.  Used in conjunction with --sample-columns")
    parser.add_argument("--sample-columns", nargs="+",
                        help="list of VCF sample (genomtype) FORMAT field names to include in variant table")
    parser.add_argument("--flanking", help="genomic region to include either side of variant", default=1000)
    parser.add_argument("--window",
                        help="initial visible window size (genomic region) in bp.  If not supplied igv.js default applies (41 bp).",
                        default=None)
    parser.add_argument("--standalone", help="embed javascript as well as data in output html", action='store_true')
    parser.add_argument("--title", help="optional title string.  Inserted into the html title tag")
    parser.add_argument("--header",
                        help="optional header html string.  Inserted into the document before the variant table")
    parser.add_argument("--footer",
                        help="optional footer html string.  Inserted into the document below the igv.js viewer")
    parser.add_argument("--sequence", help="Column of sequence (chromosome) name.  For tab-delimited sites file.",
                        default=None)
    parser.add_argument("--begin", help="Column of start position.  For tab-delimited sites file.", default=None)
    parser.add_argument("--end", help="column of end position. For tab-delimited sites file.", default=None)
    parser.add_argument("--zero_based",
                        help="Specify that the position in the data file is 0-based (e.g. UCSC files) rather than 1-based.",
                        default=None)
    parser.add_argument("--idlink", type=str, help="url link template for the VCF ID column")
    parser.add_argument("--exclude-flags", type=int,
                        help="Passed to samtools to filter alignments.  For BAM and CRAM files.", default=1536)
    parser.add_argument("--no-embed", help="Do not embed fasta or track data.  This is not common", action="store_true")
    parser.add_argument("--subsample", type=float,
                        help="Subsample bam files, keeping fraction of input alignments as indicated by input value in the range of 0.0 - 1.0")
    parser.add_argument("--maxlen", type=int, default=10000,
                        help="Maximum length of variant for single  view. Variants exceeding this lenght will be presented in split-screen (multilocus) view")
    parser.add_argument("--translate-sequence-track", help="Three-frame Translate sequence track", action="store_true")
    parser.add_argument("--tabulator", help="Enable Tabulator table with advanced filtering", action="store_true")
    parser.add_argument("--filter-config", help="YAML configuration file for column-specific filtering")
    parser.add_argument("--merge-overlaps", help="Merge overlapping regions for multi-locus features (e.g. bedpe)", action="store_true")

    args = parser.parse_args()

    # For backward compatibility with 'fasta' positional argument
    if args.fasta_ is not None and args.fasta is None:
        args.fasta = args.fasta_
    create_report(args)


if __name__ == "__main__":
    main()



def get_chromosome_alias(name):
    """
    Returns an alias for a genomic chromosome name.

    Rules:
    - If the name starts with "chr", strip the prefix.
    - If the name does not start with "chr", add the prefix.
    - If the name is "MT", return "chrM".
    - If the name is "chrM", return "MT".

    Args:
        name (str): The chromosome name.

    Returns:
        str: The alias for the chromosome name.
    """
    if name == "MT":
        return "chrM"
    elif name == "chrM":
        return "MT"
    elif name.startswith("chr"):
        return name[3:]
    else:
        return f"chr{name}"



import unittest
import pathlib
from igv_reports import fasta
from igv_reports.fasta import FastaReader
from igv_reports.twobit import TwoBitReader


class TwobitTest(unittest.TestCase):



    def test_region(self):

        expected = 'NNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNNACTCTATCTATCTATCTATCTATCTTTTTCCCCCCGGGGGGagagagagactc'
        reader = TwoBitReader('data/twobit/foo.2bit')
        data = reader.slice({
            "chr": "chr1",
            "start": 5 + 1,     # 1-based start
            "end": 100
        })

        self.assertTrue(data)
        self.assertEqual(expected, data)

    def test_remote(self):

        reader = FastaReader("https://igv.org/genomes/data/hg38/hg38.fa")
        data = reader.slice({"chr": "chr1",
                               "start": 100,
                               "end": 200})
        self.assertTrue(data)

        size = reader.get_reference_length("chr1")
        self.assertEqual(248956422, size)



    def test_remote_chrom(self):

        data = fasta.get_data("https://igv.org/genomes/data/hg38/hg38.fa",
                              {"chr": "chr5",
                               "start": 474979,
                               "end": 474998})
        self.assertEqual("ATCAGGCGGCAGAAGGTGCC", data)



import json
import pysam
import html
from .feature import Feature




class VariantTable:

    '''
    Class represents the variant selection table backed by a VCF file.
    '''

    def __init__(self, vcfFile, args):

        vcf = pysam.VariantFile(vcfFile)

       # args.info_columns, args.info_columns_prefixes, args.samples,
      #  args.sample_columns, args.idlink

        self.info_fields = args.info_columns or []
        self.idlink = args.idlink
        self.sample_fields = args.sample_columns or []
        self.variants = []
        self.features = []  # Bed-like features

        self.chr2_present = False  # Until proven otherwise
        self.chr2_dict = {}

        if args.info_columns_prefixes is not None:
            for p in args.info_columns_prefixes:
                for info in vcf.header.info:
                    if info.startswith(p):
                        self.info_fields.append(info)

        if args.samples is not None:
            self.samples = args.samples
        else :
            self.samples = []
            for s in vcf.header.samples:
                self.samples.append(s)

        for unique_id, var in enumerate(vcf.fetch()):
            self.variants.append((var, unique_id))
            chr = var.chrom
            start = var.start
            end = var.stop

            # Check for large variant (i.e. large deletion).  If too large use multilocus view
            multilocus = (end - start) > args.maxlen

            # Check for inter-chr structural variant
            chr2 = chr
            pos2 = end
            if "CHR2" in var.info:
                info = parse_info_fields(str(var))
                if "END" in info:
                    pos2 = int(info["END"])  # This is an out-of-spec use of "END", but commonly used for inter-chr SVs
                    chr2 = var.info["CHR2"]
                    if chr2 != chr:
                        self.chr2_dict[str(var)] = (chr2, pos2)
                        self.chr2_present = True
                        multilocus = True

            if multilocus:
                self.features.append((Feature(chr, start, start + 1, '', '', chr2, pos2-1, pos2), unique_id))
            else:
                self.features.append((Feature(chr, start, end, ''), unique_id))

    def to_JSON(self):

        json_array = [];

        for variant, unique_id in self.variants:

            escaped_alts = []
            for alt in variant.alts:
                escaped_alts.append(html.escape(alt))

            if self.chr2_present:
                var_key = str(variant)
                obj = {
                    'unique_id': unique_id,
                    'CHROM': variant.chrom,
                    'POSITION': variant.pos,
                    'CHROM2': self.chr2_dict[var_key][0] if var_key in self.chr2_dict else '',
                    'POSITION2': self.chr2_dict[var_key][1] if var_key in self.chr2_dict else '',
                    'REF': html.escape(variant.ref),
                    'ALT': ','.join(escaped_alts),
                    'ID': ''
                }
            else:
                obj = {
                    'unique_id': unique_id,
                    'CHROM': variant.chrom,
                    'POSITION': variant.pos,
                    'REF': html.escape(variant.ref),
                    'ALT': ','.join(escaped_alts),
                    'ID': ''
                }

            if variant.id is not None:
                obj['ID'] = render_ids(variant.id, self.idlink)

            for h in self.info_fields:
                if h in variant.info:
                    if h == 'ANN':
                        genes, effects, impacts, transcript, gene_id, aa_alt, nt_alt = decode_ann(variant)
                        obj['GENE'] = genes
                        obj['EFFECTS'] = effects
                        obj['IMPACT'] = impacts
                        obj['TRANSCRIPT'] = transcript
                        obj['GENE_ID'] = gene_id
                        obj['PROTEIN ALTERATION'] = aa_alt
                        obj['DNA ALTERATION'] = nt_alt
                    elif h == 'COSMIC_ID':
                        cid = variant.info[h]
                        if cid is not None:
                            if isinstance(cid, str):
                                return render_id([cid, self.idlink])
                            elif len(cid) == 1:
                                obj[h] = render_id([cid[0], self.idlink])
                            else:
                                tmp = ''
                                for c in cid:
                                    tmp = tmp + render_id([c, self.idlink]) + ','
                                obj[h] = tmp
                    else:
                        obj[h] = render_values(variant.info[h])
                else:
                    obj[h] = ''

            for h in self.sample_fields:
                for sample, values in variant.samples.items():
                    v = ''
                    try:
                        v = values[h]
                    except KeyError:
                        # ignore if key is not present
                        pass

                    obj[f'{sample}:{h}'] = render_values(v)

            json_array.append(obj)

        if not any('ID' in obj for obj in json_array):
            # Remove ID column if none of the records actually had an ID.
            for obj in json_array:
                del obj['ID']

        normalized_json = self.normalize_json(json_array)
        return json.dumps(normalized_json)


    def normalize_json(self, json_array):

        headers =  ['unique_id', 'CHROM', 'POSITION', 'CHROM2', 'POSITION2', 'REF', 'ALT', 'ID'] if self.chr2_present else ['unique_id', 'CHROM', 'POSITION', 'REF', 'ALT', 'ID']
        if self.info_fields is not None:
            for h in self.info_fields:
                if h == 'ANN':
                    headers = headers + ['GENE', 'EFFECTS', 'IMPACT', 'TRANSCRIPT', 'GENE_ID', 'PROTEIN ALTERATION',
                                         'DNA ALTERATION']
                else:
                    headers.append(h)

        if self.samples is not None and self.sample_fields is not None:
            for s in self.samples:
                for f in self.sample_fields:
                    h = f'{s}:{f}'
                    headers.append(h)

        rows = []
        for json in json_array:
            r = []
            for h in headers:
                if h in json:
                    r.append(json[h])
                else:
                    r.append("")
            rows.append(r)

        return {
            "headers": headers,
            "rows": rows
        }


def render_value(v):
    """Render given value to string."""
    if v is None:
        return ""
    elif isinstance(v, float):
        # ensure that we don't waste space by insignificant digits
        return f'{v:.2g}'
    elif isinstance(v, str):
        str_val = v.replace('"', '')

        if str_val.startswith('http://') or str_val.startswith('https://'):
            return create_link(str_val)

    return html.escape(str(v))


def render_values(v):
    if v is None:
        return ""

    if isinstance(v, str) or isinstance(v, int) or isinstance(v, float):
        return render_value(v)
    return ','.join(map(render_value, v))


def render_id(t):
    v, idlink = t
    if idlink is not None:
        url = idlink.replace("$$", v)
        return (f'<a href = "{url}" target="_blank">{v}</a>')
    elif v.startswith('COSM'):
        return (f'<a href = "https://cancer.sanger.ac.uk/cosmic/mutation/overview?id={v[4:]}" target="_blank">{v}</a>')
    elif v.startswith('rs'):
        return (f'<a href = "https://www.ncbi.nlm.nih.gov/snp/?term={v}" target="_blank">{v}</a>')
    return v


def render_ids(v, idlink=None):
    return ','.join(map(render_id, [(val, idlink) for val in v.split(';')]))


def decode_ann(variant):
    """
    Decode the standardized ANN field to something human readable. This is somewhat fragile and depends on the
    specifics of the ANNOVAR annotation output
    """
    annotations = ([variant.info['ANN'].split('|')] if isinstance(variant.info['ANN'],str) else [e.split('|') for e in variant.info['ANN']])
    genes = []
    effects = []
    impacts = []
    transcripts = []
    gene_ids = []
    aa_alts = []
    nt_alts = []
    for allele in variant.alts:
        for ann in annotations:
            ann_allele, kind, impact, gene, gene_id = ann[:5]
            feature_id = ann[6]
            nt_mod, aa_mod = ann[9:11]

            if allele != ann_allele:
                continue

            genes.append(gene)
            gene_ids.append(gene_id)
            effects.append(kind.replace('&', '/'))
            impacts.append(impact)
            transcripts.append(feature_id)
            aa_alts.append(aa_mod)
            nt_alts.append(nt_mod)
    return '<br>'.join(genes), '<br>'.join(effects), '<br>'.join(impacts), '<br>'.join(transcripts), \
           '<br>'.join(gene_ids), '<br>'.join(aa_alts), '<br>'.join(nt_alts)


def create_link(url):
    """Create an html link for the given url"""
    return (f'<a href = "{url}" target="_blank">{url}</a>')


#1	564466	26582	N	<TRA>	.	PASS	PRECISE;SVMETHOD=Snifflesv1.0.2;CHR2=MT;END=3916;STD_quant_start=198.695999;STD_quant_stop=234.736235;Kurtosis_quant_start=0.913054;Kurtosis_quant_stop=-0.183504;SVTYPE=TRA;SUPTYPE=SR;SVLEN=-1199826434;STRANDS=-+;STRANDS2=2,9,2,9;RE=11	GT:DR:DV	./.:.:11
def parse_info_fields(record):
    tokens = record.split('\t')
    info_fields = tokens[7].split(';')
    info = {}
    for ifield in info_fields:
        key_value = ifield.split("=")
        if len(key_value) == 2:
            info[key_value[0]] = key_value[1]
    return info




import pysam
from igv_reports.chralias import build_aliastable

class BamReader:

    def __init__(self, filetype, filename, args = None):
        self.filtetype = filetype
        self.filename = filename
        self.args = args

        samargs = ["-H", filename]
        header = pysam.view(*samargs)
        seqnames = parse_seqnames(header)
        self.aliastable = build_aliastable(seqnames)

    # sam flag supports unit tests
    def slice(self, region=None, region2=None, sam = False):

        samargs = ["-h", self.filename] if sam else ["-b", "-h", self.filename]

        if self.args is not None and self.args.subsample is not None:
            samargs.append("--subsample")
            samargs.append(str(self.args.subsample))

        if self.filtetype == "cram" and self.args is not None:
            if self.args.fasta is None:
                raise ValueError("Cram file requires a reference fasta file")
            samargs.append("-T")
            samargs.append(self.args.fasta)

        if self.args is not None and hasattr(self.args, "exclude_flags"):
            if self.args.exclude_flags != 0:
                samargs.append("-F")
                samargs.append(str(self.args.exclude_flags))
        else:
            samargs.append("-F")
            samargs.append("1536")

        if region:
            if region['start'] is not None:
                range_string = self.get_chrname(region['chr']) + ":" + str(region['start']) + "-" + str(region['end'])
            else:
                range_string = self.get_chrname(region['chr'])
            samargs.append(range_string)
        if region2:
            range_string = self.get_chrname(region2['chr']) + ":" + str(region2['start']) + "-" + str(region2['end'])
            samargs.append(range_string)

        return pysam.view(*samargs)

    def get_chrname(self, c):
        if c in self.aliastable:
            return self.aliastable[c]
        else:
            return c



def parse_seqnames(header):
    seqnames = []
    lines = header.split('\n')
    for line in lines:
        if line.startswith('@SQ'):
            idx1 = line.find("SN:")
            if idx1 > 0:
                idx2 = line.find("\t", idx1)
                seqnames.append(line[idx1+3:idx2])

    return seqnames



from . import regions
import pysam

from .chr_alias import get_chromosome_alias


def get_data(fasta_file,region=None):

    if None == region:

        with open(fasta_file,"r") as f:
            return(f.read())

    else :

        if isinstance(region,str):
            region = regions.parse_region(region)

        chr = region["chr"]
        start = region["start"] - 1
        end = region["end"]

        fasta = pysam.FastaFile(fasta_file)

        slice_seq = fasta.fetch(chr, start, end)

        fasta.close()

        return slice_seq


class FastaReader:

    def __init__(self, file):
        self.fasta = pysam.FastaFile(file)

    def slice(self, region = None):

        if None == region:
            with open(self.fasta,"r") as f:
                return(f.read())

        else :

            if isinstance(region,str):
                region = regions.parse_region(region)

            chr = region["chr"]
            start = region["start"] - 1 if region["start"] is not None else None
            end = region["end"] if region["end"] is not None else None

            try:
                seq = self.fasta.fetch(chr, start, end)


            except KeyError:
                chr = get_chromosome_alias(chr)
                seq = self.fasta.fetch(chr, start, end)

            return seq

    def get_reference_length(self, chr):
        """
        Returns the length of the reference sequence.
        """
        try:
            return self.fasta.get_reference_length(chr)

        except KeyError:
            chr = get_chromosome_alias(chr)
            return self.fasta.get_reference_length(chr)





import gzip
import io
import requests

def getstream(file):
    # TODO -- gcs

    if file.startswith('http://') or file.startswith('https://'):
        response = requests.get(file)
        status_code = response.status_code  # TODO Do something with this

        if file.endswith('.gz'):
            content = response.content
            text = gzip.decompress(content).decode('utf-8')
        else:
            text = response.text
        f = io.StringIO(text)
        return f

    elif file.endswith('.gz'):
        f = gzip.open(file, mode='rt')

    else:
        f = open(file, encoding='UTF-8')

    return f


# Check if the http or https URL resource exists.
# Used to probe for index files

def resource_exists(url):

    try:
        response = requests.get(url)
        status_code = response.status_code
        return status_code >= 200 and status_code < 400
    except:
        return False




from . import regions
from pytwobit import TwoBit

from .chr_alias import get_chromosome_alias


class TwoBitReader:

    def __init__(self, path):
        self.path = path
        self.twobit = TwoBit(path)

    def slice(self, region):

        if isinstance(region,str):
            region = regions.parse_region(region)

        chr = region["chr"]
        start = None
        if region["start"] is not None:
            start = region["start"] - 1
        end = region["end"]

        data = self.twobit.fetch(chr, start, end)

        if data is None:
            chr = get_chromosome_alias(chr)
            data = self.twobit.fetch(chr, start, end)

        return data


    def get_reference_length(self, chr):
        """
        Get the length of the reference sequence for a given chromosome.

        Args:
            chr (str): The chromosome name.

        Returns:
            int: The length of the reference sequence.
        """
        length = self.twobit.get_reference_length(chr)

        if length is None:
            chr = get_chromosome_alias(chr)
            length = self.twobit.get_reference_length(chr)

        return length



